using System;
using System.Drawing;
using System.Numerics;
using SlatedGameToolkit.Framework.Graphics.OpenGL;
using SlatedGameToolkit.Framework.Graphics.Render.Programs;
using SlatedGameToolkit.Framework.Graphics.Render.Shaders;
using SlatedGameToolkit.Framework.Graphics.Textures;
using SlatedGameToolkit.Framework.Graphics.Window;
using SlatedGameToolkit.Framework.Utilities;

namespace SlatedGameToolkit.Framework.Graphics.Render
{

    /// <summary>
    /// Represents a grouping of similar meshes that can be rendered with similar properties.
    /// Specifically, properties that need to be the same are:
    /// Texture and it's respective properties, model matrix, and vertex buffer array and objects.
    /// </summary>
    public class MeshBatchRenderer : IDisposable {
        public bool Debug { get; set; }
        private const int VERTEX_LENGTH = 9;
        private bool disposed;
        private float batchDelta;
        public GLContext GLContext {get; private set; }
        private int projULoc, viewULoc, modelULoc, texturedULoc, singleChanneledULoc, flippedULoc;
        private Camera camera;
        private RenderProgram renderProgram;
        private ITexture texture;
        private VertexArrayBuffers vertexBuffers;
        public bool Batching { get; private set; }

        private Matrix4x4 modelsMatrix;
        private float[] data;
        private uint[] indices;
        private int[] lengths;
        private int[] indiceOffsets;
        private int[] verticeOffsets;
        private int dataIndex, indicesIndex, offsetIndex;

        /// <summary>
        /// Instantiates a mesh batch with a default vertex and fragment shader.
        /// </summary>
        /// <param name="camera">The camera to use for providing the correct matrices.</param>
        /// <param name="BatchVertexSize">The vertex batch size. Defaults to 1024. This is multiplied by 9 for each individual value of the vertex.</param>
        /// <param name="glContext">The glContext to associate the shaders with. Defaults to the currently active one.</param>
        public MeshBatchRenderer(Camera camera, uint BatchVertexSize = 1024, GLContext glContext = null) {
            VertexShader vert = new VertexShader(EmbeddedResUtils.ReadEmbeddedResourceText("default.vert"));
            FragmentShader frag = new FragmentShader(EmbeddedResUtils.ReadEmbeddedResourceText("default.frag"));
            this.GLContext = glContext ?? WindowContextsManager.CurrentGL;
            renderProgram = new RenderProgram(this.GLContext, vert, frag);

            this.camera = camera ?? throw new ArgumentNullException("camera");
            if (BatchVertexSize < 1) throw new ArgumentException("Batch vertex size cannot be less than 1.");
            indices = new uint[BatchVertexSize];
            lengths = new int[BatchVertexSize];
            indiceOffsets = new int[BatchVertexSize];
            data = new float[BatchVertexSize * VERTEX_LENGTH];
            verticeOffsets = new int[BatchVertexSize];
            vertexBuffers = new VertexArrayBuffers(GLContext);
            VertexAttributeDefinition[] definitions = new VertexAttributeDefinition[3];
            definitions[0] = new VertexAttributeDefinition((uint)GLContext.GetAttribLocation(renderProgram.Handle, "aPosition"), 3);
            definitions[1] = new VertexAttributeDefinition((uint)GLContext.GetAttribLocation(renderProgram.Handle, "aColor"), 4);
            definitions[2] = new VertexAttributeDefinition((uint)GLContext.GetAttribLocation(renderProgram.Handle, "aTexCoord"), 2);
            vertexBuffers.defineVertexAttributes(definitions: definitions);
            renderProgram.Use();
            modelULoc = GLContext.GetUniformLocation(renderProgram.Handle, "models");
            viewULoc = GLContext.GetUniformLocation(renderProgram.Handle, "view");
            projULoc = GLContext.GetUniformLocation(renderProgram.Handle, "projection");
            texturedULoc = GLContext.GetUniformLocation(renderProgram.Handle, "textured");
            singleChanneledULoc = GLContext.GetUniformLocation(renderProgram.Handle, "singleChanneled");
            flippedULoc = GLContext.GetUniformLocation(renderProgram.Handle, "flipped");

            GLContext.UniformMatrix4fv(projULoc, 1, false, camera.ProjectionMatrix.ToColumnMajorArray());
        }

        /// <summary>
        /// Instantiates the mesh batch with a custom shader program.
        /// 
        /// For this shader to work with this mesh batch, it must have at least defined the required attributes and uniforms.
        /// </summary>
        /// <param name="aPosition">The position vertex attribute's name. A vec3. To be used to detect the index of the attribute in the shader program.</param>
        /// <param name="aColor">The color vertex attribute's name. A vec4. To be used to detect the index of the attribute in the shader program.</param>
        /// <param name="aTexCoord">The texture coordinate vertex attribute's name. A vec2. To be used to detect the index of the attribute in the shader program.</param>
        /// <param name="uModels">The models 4x4 matrix's uniform name. To be used to detect the index of the uniform in the shader program.</param>
        /// <param name="uView">The view 4x4 matrix's uniform name. To be used to detect the index of the uniform in the shader program.</param>
        /// <param name="uProjection">The projection 4x4 matrix's uniform name. To be used to detect the index of the uniform in the shader program.</param>
        /// <param name="uTextured">The name of a boolean uniform dictating whether or not the batch of meshes have an associated texture.</param>
        /// <param name="uSingleChanneled">The name of a boolean uniform dictating whether or not the current texture is single channelled and only uses red.</param>
        /// <param name="uFlipped">The name of a boolean uniform dictating whether or not the current texture is to have it's coordinates flipped.</param>
        /// <param name="camera">The camera to use to provide the matrices.</param>
        /// <param name="renderProgram">The render program to use for the actual rendering.</param>
        /// <param name="BatchVertexSize">The vertex buffer size. Defaults to 1024, and is multiplied by 9 for the individual values of each vertex.</param>
        public MeshBatchRenderer(string aPosition, string aColor, string aTexCoord, string uModels, string uView, string uProjection, string uTextured, string uSingleChanneled, string uFlipped, Camera camera, RenderProgram renderProgram, uint BatchVertexSize = 1024) {
            this.renderProgram = renderProgram ?? throw new ArgumentNullException("renderProgram");
            this.camera = camera ?? throw new ArgumentNullException("camera");
            if (BatchVertexSize < 1) throw new ArgumentException("Batch vertex size cannot be less than 1.");
            this.GLContext = renderProgram.GLContext;

            indices = new uint[BatchVertexSize];
            lengths = new int[BatchVertexSize];
            indiceOffsets = new int[BatchVertexSize];
            data = new float[BatchVertexSize * VERTEX_LENGTH];
            verticeOffsets = new int[BatchVertexSize];
            vertexBuffers = new VertexArrayBuffers(GLContext);
            VertexAttributeDefinition[] definitions = new VertexAttributeDefinition[3];
            definitions[0] = new VertexAttributeDefinition((uint)GLContext.GetAttribLocation(renderProgram.Handle, aPosition), 3);
            definitions[1] = new VertexAttributeDefinition((uint)GLContext.GetAttribLocation(renderProgram.Handle, aColor), 4);
            definitions[2] = new VertexAttributeDefinition((uint)GLContext.GetAttribLocation(renderProgram.Handle, aTexCoord), 2);

            renderProgram.Use();
            modelULoc = GLContext.GetUniformLocation(renderProgram.Handle, uModels);
            viewULoc = GLContext.GetUniformLocation(renderProgram.Handle, uView);
            projULoc = GLContext.GetUniformLocation(renderProgram.Handle, uProjection);
            texturedULoc = GLContext.GetUniformLocation(renderProgram.Handle, uTextured);
            singleChanneledULoc = GLContext.GetUniformLocation(renderProgram.Handle, uSingleChanneled);
            flippedULoc = GLContext.GetUniformLocation(renderProgram.Handle, uFlipped);

            GLContext.UniformMatrix4fv(projULoc, 1, false, camera.ProjectionMatrix.ToColumnMajorArray());
        }

        /// <summary>
        /// Begins a mesh batch for rendering.
        /// </summary>
        /// <param name="modelsMatrix">The models matrix.</param>
        /// <param name="delta">The time elapsed since the last update.</param>
        public virtual void Begin(Matrix4x4 modelsMatrix, double delta) {
            if (Batching) throw new InvalidOperationException("This batch is already started.");
            this.Batching = true;
            this.modelsMatrix = modelsMatrix;

            this.batchDelta = (float)delta;
        }

        /// <summary>
        /// Draws the given mesh.
        /// </summary>
        /// <param name="mesh">Draws the given mesh.</param>
        public virtual void Draw(IMesh mesh) {
            if (!Batching) throw new InvalidOperationException("This batch has not been begun.");
            IPositionInterpolable moveable = mesh as IPositionInterpolable;
            if (moveable != null) {
                moveable.InterpolatePosition(batchDelta);
            }
            if (mesh.Texture?.Handle != this.texture?.Handle) {
                Flush();
                this.texture = mesh.Texture;
                GLContext.Uniform1i(texturedULoc, texture == null ? 0 : 1);
                if (texture != null) {
                    GLContext.Uniform1i(flippedULoc, texture.Flipped ? 1 : 0);
                    GLContext.Uniform1i(singleChanneledULoc, texture.SingleChanneled ? 1 : 0);
                }
            }
            ValueTuple<Vector3, Vector2>[] vertices = mesh.Vertices;
            uint[] indices = mesh.Elements;
            if (vertices.Length * VERTEX_LENGTH + dataIndex >= data.Length || indices.Length + indicesIndex >= this.indices.Length) 
                Flush();
            for (int i = 0; i < vertices.Length; i++) {
                data[dataIndex] = vertices[i].Item1.X;
                dataIndex++;
                data[dataIndex] = vertices[i].Item1.Y;
                dataIndex++;
                data[dataIndex] = vertices[i].Item1.Z;
                dataIndex++;

                data[dataIndex] = mesh.Color.RedAsFloat();
                dataIndex++;
                data[dataIndex] = mesh.Color.GreenAsFloat();
                dataIndex++;
                data[dataIndex] = mesh.Color.BlueAsFloat();
                dataIndex++;
                data[dataIndex] = mesh.Color.AlphaAsFloat();
                dataIndex++;

                data[dataIndex] = vertices[i].Item2.X;
                dataIndex++;
                data[dataIndex] = vertices[i].Item2.Y;
                dataIndex++;
            }
            int elementsCount = indices.Length;
            Array.Copy(indices, 0, this.indices, indicesIndex, elementsCount);
            indiceOffsets[offsetIndex] = indicesIndex * sizeof(uint);
            indicesIndex += elementsCount;
            if (offsetIndex + 1 < verticeOffsets.Length) {
                verticeOffsets[offsetIndex + 1] = verticeOffsets[offsetIndex] + vertices.Length;
            }
            lengths[offsetIndex] = elementsCount;
            offsetIndex++;
        }
        
        public virtual void End() {
            if (!Batching) throw new InvalidOperationException("This batch was never started.");
            Flush();
            this.Batching = false;
        }

        public virtual void Flush() {
            if (!Batching) throw new InvalidOperationException("This batch has not been begun.");
            if (offsetIndex == 0) return;
            if (texture != null) {
                GLContext.BindTexture(TextureTarget.Texture2D, texture.Handle);
            }
            renderProgram.Use();
            camera.InterpolatePosition(batchDelta);

            vertexBuffers.BufferVertices(data, true);
            vertexBuffers.BufferIndices(indices, true);
            GLContext.UniformMatrix4fv(modelULoc, 1, false, modelsMatrix.ToColumnMajorArray());
            if (camera.ViewChanged) GLContext.UniformMatrix4fv(viewULoc, 1, false, camera.ViewMatrix.ToColumnMajorArray());
            if (camera.ProjectionChanged) GLContext.UniformMatrix4fv(projULoc, 1, false, camera.ProjectionMatrix.ToColumnMajorArray());

            if (Debug) {
                GLContext.PolygonMode(MaterialFace.FrontAndBack, PolygonMode.Line);
            } else {
                GLContext.PolygonMode(MaterialFace.FrontAndBack, PolygonMode.Fill);
            }
            GLContext.MultiDrawElementsBaseVertex(PrimitiveType.Triangles, lengths, DrawElementsType.UnsignedInt, indiceOffsets, offsetIndex, verticeOffsets);
            dataIndex = 0;
            indicesIndex = 0;
            offsetIndex = 0;
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing)
                {
                }
                if (disposed) return;
                if (Batching) End();
                disposed = true;
                vertexBuffers.Dispose();
                renderProgram.Dispose();
                data = null;
                indices = null;
                lengths = null;
                verticeOffsets = null;
            }
        }

        ~MeshBatchRenderer()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
    }
}